%  @copyright 2010-2011 Zuse Institute Berlin

%   Licensed under the Apache License, Version 2.0 (the "License");
%   you may not use this file except in compliance with the License.
%   You may obtain a copy of the License at
%
%       http://www.apache.org/licenses/LICENSE-2.0
%
%   Unless required by applicable law or agreed to in writing, software
%   distributed under the License is distributed on an "AS IS" BASIS,
%   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%   See the License for the specific language governing permissions and
%   limitations under the License.

%% @author Nico Kruber <kruber@zib.de>
%% @doc    Unit tests for src/nodelist.erl
%% @end
%% @version $Id$
-module(nodelist_SUITE).

-author('kruber@zib.de').
-vsn('$Id$').

-compile(export_all).

-include("unittest.hrl").
-include("scalaris.hrl").

-dialyzer({[no_opaque, no_return], [tester_largest_smaller_than2/1,
                                    tester_largest_smaller_than3/1]}).

all() ->
    [test_new, test_mk2, test_mk4, test_trunc,
     test_remove2, test_remove3, test_filter2, test_filter3,
     test_filter_min_length4, test_merge, test_add_node, test_add_nodes,
     test_update_ids, test_to_list, test_lupdate_ids, test_lremove_outdated,
     tester_largest_smaller_than2,
     tester_largest_smaller_than3,
     test_succ_ord_id].

suite() -> [ {timetrap, {seconds, 40}} ].

init_per_suite(Config) ->
    unittest_helper:start_minimal_procs(Config, [], true).

end_per_suite(Config) ->
    _ = unittest_helper:stop_minimal_procs(Config),
    ok.

% TODO: add tests for simple node lists (the following tests only test functions with neighborhood structures)

-define(PID(Nr), comm:make_global(list_to_pid("<0.0." ++ erlang:integer_to_list(Nr) ++ ">"))).
-define(KEY(Nr), rt_SUITE:number_to_key(Nr)).

% use macro instead of function so that the output in case of errors is better:
-define(compare_neighborhood(Neighborhood, Node, Pred, Preds, Succ, Succs, RealPred, RealSucc),
        compare_neighborhood(??Neighborhood, Neighborhood, Node, Pred, Preds, Succ, Succs, RealPred, RealSucc)).

-spec compare_neighborhood(NeighborhoodStr::string(), N::nodelist:neighborhood(), Node::node:node_type(), Pred::node:node_type(), Preds::nodelist:snodelist(), Succ::node:node_type(), Succs::nodelist:snodelist(), RealPred::boolean(), RealSucc::boolean()) -> ok.
compare_neighborhood(NeighborhoodStr, N, Node, Pred, Preds, Succ, Succs, RealPred, RealSucc) ->
    ?equals_w_note(nodelist:node(N), Node, io_lib:format("N = ~s", [NeighborhoodStr])),
    ?equals_w_note(nodelist:nodeid(N), node:id(Node), io_lib:format("N = ~s", [NeighborhoodStr])),
    ?equals_w_note(nodelist:pred(N), Pred, io_lib:format("N = ~s", [NeighborhoodStr])),
    ?equals_w_note(nodelist:preds(N), Preds, io_lib:format("N = ~s", [NeighborhoodStr])),
    ?equals_w_note(nodelist:succ(N), Succ, io_lib:format("N = ~s", [NeighborhoodStr])),
    ?equals_w_note(nodelist:succs(N), Succs, io_lib:format("N = ~s", [NeighborhoodStr])),
    ?equals_w_note(nodelist:has_real_pred(N), RealPred, io_lib:format("N = ~s", [NeighborhoodStr])),
    ?equals_w_note(nodelist:has_real_succ(N), RealSucc, io_lib:format("N = ~s", [NeighborhoodStr])),
    ok.

%% @doc Tests neighborhood creation with new_neighborhood and compares the
%%      getters with what is expected.
-spec test_new(any()) -> ok.
test_new(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N15 = node:new(?PID(2), ?KEY(15), 1), % N14 updated its ID to 15 here
    N33 = node:new(?PID(3), ?KEY(33), 0),
    
    Neighb11 = nodelist:new_neighborhood(N8),
    ?compare_neighborhood(Neighb11, N8, N8, [N8], N8, [N8], false, false),
    
    Neighb21 = nodelist:new_neighborhood(N8, N14),
    Neighb22 = nodelist:new_neighborhood(N14, N8),
    Neighb23 = nodelist:new_neighborhood(N8, N8),
    Neighb24 = nodelist:new_neighborhood(N15, N14),
    ?compare_neighborhood(Neighb21, N8, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(Neighb22, N14, N8, [N8], N8, [N8], true, true),
    ?compare_neighborhood(Neighb23, N8, N8, [N8], N8, [N8], false, false),
    ?compare_neighborhood(Neighb24, N15, N15, [N15], N15, [N15], false, false),
    
    Neighb31 = nodelist:new_neighborhood(N14, N8,  N33),
    Neighb32 = nodelist:new_neighborhood(N33, N8,  N14),
    Neighb33 = nodelist:new_neighborhood(N8,  N14, N33),
    Neighb34 = nodelist:new_neighborhood(N33, N14, N8),
    Neighb35 = nodelist:new_neighborhood(N14, N33, N8),
    Neighb36 = nodelist:new_neighborhood(N8,  N33, N14),
    Neighb37 = nodelist:new_neighborhood(N14, N8,  N8),
    Neighb38 = nodelist:new_neighborhood(N8, N8,  N8),
    Neighb39 = nodelist:new_neighborhood(N14, N8,  N15),
    Neighb310 = nodelist:new_neighborhood(N15, N8,  N14),
    Neighb311 = nodelist:new_neighborhood(N8, N15,  N14),
    Neighb312 = nodelist:new_neighborhood(N14, N15,  N8),
    ?compare_neighborhood(Neighb31,  N8,  N33, [N33], N14, [N14], true, true),
    ?compare_neighborhood(Neighb32,  N8,  N33, [N33], N14, [N14], true, true),
    ?compare_neighborhood(Neighb33,  N14, N8,  [N8],  N33, [N33], true, true),
    ?compare_neighborhood(Neighb34,  N14, N8,  [N8],  N33, [N33], true, true),
    ?compare_neighborhood(Neighb35,  N33, N14, [N14], N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb36,  N33, N14, [N14], N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb37,  N8,  N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(Neighb38,  N8,  N8,  [N8],  N8,  [N8],  false, false),
    ?compare_neighborhood(Neighb39,  N8,  N15, [N15], N15, [N15], true, true),
    ?compare_neighborhood(Neighb310, N8,  N15, [N15], N15, [N15], true, true),
    ?compare_neighborhood(Neighb311, N15, N8,  [N8],  N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb312, N15, N8,  [N8],  N8,  [N8],  true, true),
    
    % tests with pred/succ nodes equal to the base node but newer:
    ?expect_exception(nodelist:new_neighborhood(N14, N15),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:new_neighborhood(N15, N14, N8),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:new_neighborhood(N8, N14, N15),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ok.

%% @doc Tests neighborhood creation with mk_neighborhood/2 and compares the
%%      getters with what is expected.
-spec test_mk2(any()) -> ok.
test_mk2(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N15 = node:new(?PID(2), ?KEY(15), 1), % N14 updated its ID to 15 here
    N33 = node:new(?PID(3), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),
    
    Neighb11 = nodelist:mk_neighborhood([N8], N8),
    Neighb12 = nodelist:mk_neighborhood([], N8),
    ?compare_neighborhood(Neighb11, N8, N8, [N8], N8, [N8], false, false),
    ?compare_neighborhood(Neighb12, N8, N8, [N8], N8, [N8], false, false),
    
    Neighb21 = nodelist:mk_neighborhood([N8, N14], N8),
    Neighb22 = nodelist:mk_neighborhood([N14, N8], N14),
    Neighb23 = nodelist:mk_neighborhood([N8, N8], N8),
    Neighb24 = nodelist:mk_neighborhood([N14], N8),
    Neighb25 = nodelist:mk_neighborhood([N8], N14),
    Neighb26 = nodelist:mk_neighborhood([N14], N15),
    ?compare_neighborhood(Neighb21, N8,  N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(Neighb22, N14, N8,  [N8],  N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb23, N8,  N8,  [N8],  N8,  [N8],  false, false),
    ?compare_neighborhood(Neighb24, N8,  N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(Neighb25, N14, N8,  [N8],  N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb26, N15, N15, [N15], N15, [N15], false, false),
    
    Neighb31 = nodelist:mk_neighborhood( [N14, N8,  N33], N8),
    Neighb32 = nodelist:mk_neighborhood( [N33, N8,  N14], N8),
    Neighb33 = nodelist:mk_neighborhood( [N8,  N14, N33], N14),
    Neighb34 = nodelist:mk_neighborhood( [N33, N14, N8],  N14),
    Neighb35 = nodelist:mk_neighborhood( [N14, N33, N8],  N33),
    Neighb36 = nodelist:mk_neighborhood( [N8,  N33, N14], N33),
    Neighb37 = nodelist:mk_neighborhood( [N14, N8,  N8],  N8),
    Neighb38 = nodelist:mk_neighborhood( [N8, N8,  N8],   N8),
    Neighb39 = nodelist:mk_neighborhood( [N14, N15],      N8),
    Neighb310 = nodelist:mk_neighborhood([N15, N14],      N8),
    Neighb311 = nodelist:mk_neighborhood([N14, N8],       N15),
    Neighb312 = nodelist:mk_neighborhood([N8, N14],       N15),
    ?compare_neighborhood(Neighb31,  N8,  N33, [N33, N14], N14, [N14, N33], true, true),
    ?compare_neighborhood(Neighb32,  N8,  N33, [N33, N14], N14, [N14, N33], true, true),
    ?compare_neighborhood(Neighb33,  N14, N8,  [N8, N33],  N33, [N33, N8],  true, true),
    ?compare_neighborhood(Neighb34,  N14, N8,  [N8, N33],  N33, [N33, N8],  true, true),
    ?compare_neighborhood(Neighb35,  N33, N14, [N14, N8],  N8,  [N8, N14],  true, true),
    ?compare_neighborhood(Neighb36,  N33, N14, [N14, N8],  N8,  [N8, N14],  true, true),
    ?compare_neighborhood(Neighb37,  N8,  N14, [N14],      N14, [N14],      true, true),
    ?compare_neighborhood(Neighb38,  N8,  N8,  [N8],       N8,  [N8],       false, false),
    ?compare_neighborhood(Neighb39,  N8,  N15, [N15],      N15, [N15],      true, true),
    ?compare_neighborhood(Neighb310, N8,  N15, [N15],      N15, [N15],      true, true),
    ?compare_neighborhood(Neighb311, N15, N8,  [N8],       N8,  [N8],       true, true),
    ?compare_neighborhood(Neighb312, N15, N8,  [N8],       N8,  [N8],       true, true),
    
    % some tests with duplicated nodes:
    Neighb41 = nodelist:mk_neighborhood( [N14, N8, N8,  N33], N8),
    Neighb42 = nodelist:mk_neighborhood( [N33, N8,  N14, N14], N8),
    Neighb43 = nodelist:mk_neighborhood( [N8,  N14, N33, N14], N14),
    Neighb44 = nodelist:mk_neighborhood( [N33, N14, N8, N33, N14],  N14),
    Neighb45 = nodelist:mk_neighborhood( [N14, N33, N8, N14, N8],  N33),
    ?compare_neighborhood(Neighb41,  N8,  N33, [N33, N14], N14, [N14, N33], true, true),
    ?compare_neighborhood(Neighb42,  N8,  N33, [N33, N14], N14, [N14, N33], true, true),
    ?compare_neighborhood(Neighb43,  N14, N8,  [N8, N33],  N33, [N33, N8],  true, true),
    ?compare_neighborhood(Neighb44,  N14, N8,  [N8, N33],  N33, [N33, N8],  true, true),
    ?compare_neighborhood(Neighb45,  N33, N14, [N14, N8],  N8,  [N8, N14],  true, true),
    
    % tests with nodes equal to the base node in the node list:
    ?compare_neighborhood(nodelist:mk_neighborhood([N8, N33, N14], N15),
                          N15, N8, [N8, N33], N33, [N33, N8], true, true),
    ?expect_exception(nodelist:mk_neighborhood([N8, N33, N15], N14),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:mk_neighborhood([N8, N15, N33], N14),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:mk_neighborhood([N15, N8, N33], N14),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:mk_neighborhood([N15], N14),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    
    % a few tests with longer lists
    Neighb51 = nodelist:mk_neighborhood([N66, N8, N50, N33, N42], N14),
    Neighb52 = nodelist:mk_neighborhood([N66, N8, N50, N14, N33, N42], N14),
    Neighb53 = nodelist:mk_neighborhood([N66, N8, N50, N14, N33, N42], N15),
    ?compare_neighborhood(Neighb51, N14, N8, [N8, N66, N50, N42, N33], N33, [N33, N42, N50, N66, N8], true, true),
    ?compare_neighborhood(Neighb52, N14, N8, [N8, N66, N50, N42, N33], N33, [N33, N42, N50, N66, N8], true, true),
    ?compare_neighborhood(Neighb53, N15, N8, [N8, N66, N50, N42, N33], N33, [N33, N42, N50, N66, N8], true, true),
    ok.

%% @doc Tests neighborhood creation with mk_neighborhood/4 and compares the
%%      getters with what is expected.
-spec test_mk4(any()) -> ok.
test_mk4(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N15 = node:new(?PID(2), ?KEY(15), 1), % N14 updated its ID to 15 here
    N33 = node:new(?PID(3), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),
    
    Neighb11 = nodelist:mk_neighborhood([N8], N8, 1, 1),
    Neighb12 = nodelist:mk_neighborhood([], N8, 1, 1),
    ?compare_neighborhood(Neighb11, N8, N8, [N8], N8, [N8], false, false),
    ?compare_neighborhood(Neighb12, N8, N8, [N8], N8, [N8], false, false),
    
    Neighb21 = nodelist:mk_neighborhood([N8, N14], N8, 1, 1),
    Neighb22 = nodelist:mk_neighborhood([N14, N8], N14, 1, 1),
    Neighb23 = nodelist:mk_neighborhood([N8, N8], N8, 1, 1),
    Neighb24 = nodelist:mk_neighborhood([N14], N8, 1, 1),
    Neighb25 = nodelist:mk_neighborhood([N8], N14, 1, 1),
    Neighb26 = nodelist:mk_neighborhood([N14], N15, 1, 1),
    ?compare_neighborhood(Neighb21, N8,  N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(Neighb22, N14, N8,  [N8],  N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb23, N8,  N8,  [N8],  N8,  [N8],  false, false),
    ?compare_neighborhood(Neighb24, N8,  N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(Neighb25, N14, N8,  [N8],  N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb26, N15, N15, [N15], N15, [N15], false, false),
    
    Neighb31 = nodelist:mk_neighborhood( [N14, N8,  N33], N8, 1, 1),
    Neighb32 = nodelist:mk_neighborhood( [N33, N8,  N14], N8, 1, 1),
    Neighb33 = nodelist:mk_neighborhood( [N8,  N14, N33], N14, 1, 1),
    Neighb34 = nodelist:mk_neighborhood( [N33, N14, N8],  N14, 1, 1),
    Neighb35 = nodelist:mk_neighborhood( [N14, N33, N8],  N33, 1, 1),
    Neighb36 = nodelist:mk_neighborhood( [N8,  N33, N14], N33, 1, 1),
    Neighb37 = nodelist:mk_neighborhood( [N14, N8,  N8],  N8, 1, 1),
    Neighb38 = nodelist:mk_neighborhood( [N8, N8,  N8],   N8, 1, 1),
    Neighb39 = nodelist:mk_neighborhood( [N14, N15],      N8, 1, 1),
    Neighb310 = nodelist:mk_neighborhood([N15, N14],      N8, 1, 1),
    Neighb311 = nodelist:mk_neighborhood([N14, N8],       N15, 1, 1),
    Neighb312 = nodelist:mk_neighborhood([N8, N14],       N15, 1, 1),
    ?compare_neighborhood(Neighb31,  N8,  N33, [N33], N14, [N14], true, true),
    ?compare_neighborhood(Neighb32,  N8,  N33, [N33], N14, [N14], true, true),
    ?compare_neighborhood(Neighb33,  N14, N8,  [N8],  N33, [N33], true, true),
    ?compare_neighborhood(Neighb34,  N14, N8,  [N8],  N33, [N33], true, true),
    ?compare_neighborhood(Neighb35,  N33, N14, [N14], N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb36,  N33, N14, [N14], N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb37,  N8,  N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(Neighb38,  N8,  N8,  [N8],  N8,  [N8],  false, false),
    ?compare_neighborhood(Neighb39,  N8,  N15, [N15], N15, [N15], true, true),
    ?compare_neighborhood(Neighb310, N8,  N15, [N15], N15, [N15], true, true),
    ?compare_neighborhood(Neighb311, N15, N8,  [N8],  N8,  [N8],  true, true),
    ?compare_neighborhood(Neighb312, N15, N8,  [N8],  N8,  [N8],  true, true),
    
    % some tests with duplicated nodes:
    Neighb41 = nodelist:mk_neighborhood( [N14, N8, N8,  N33], N8, 1, 1),
    Neighb42 = nodelist:mk_neighborhood( [N33, N8,  N14, N14], N8, 1, 1),
    Neighb43 = nodelist:mk_neighborhood( [N8,  N14, N33, N14], N14, 1, 1),
    Neighb44 = nodelist:mk_neighborhood( [N33, N14, N8, N33, N14],  N14, 1, 1),
    Neighb45 = nodelist:mk_neighborhood( [N14, N33, N8, N14, N8],  N33, 1, 1),
    ?compare_neighborhood(Neighb41,  N8,  N33, [N33], N14, [N14], true, true),
    ?compare_neighborhood(Neighb42,  N8,  N33, [N33], N14, [N14], true, true),
    ?compare_neighborhood(Neighb43,  N14, N8,  [N8],  N33, [N33], true, true),
    ?compare_neighborhood(Neighb44,  N14, N8,  [N8],  N33, [N33], true, true),
    ?compare_neighborhood(Neighb45,  N33, N14, [N14], N8,  [N8],  true, true),
    
    % tests with nodes equal to the base node in the node list:
    ?compare_neighborhood(nodelist:mk_neighborhood([N8, N33, N14], N15, 3, 3),
                          N15, N8, [N8, N33], N33, [N33, N8], true, true),
    ?expect_exception(nodelist:mk_neighborhood([N8, N33, N15], N14, 3, 3),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:mk_neighborhood([N8, N15, N33], N14, 3, 3),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:mk_neighborhood([N15, N8, N33], N14, 3, 3),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:mk_neighborhood([N15], N14, 3, 3),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),

    ?compare_neighborhood(nodelist:mk_neighborhood([N8, N33, N14], N15, 1, 1),
                          N15, N8, [N8], N33, [N33], true, true),
    ?expect_exception(nodelist:mk_neighborhood([N8, N33, N15], N14, 1, 1),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:mk_neighborhood([N8, N15, N33], N14, 1, 1),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:mk_neighborhood([N15, N8, N33], N14, 1, 1),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:mk_neighborhood([N15], N14, 1, 1),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),

    % a few tests with longer lists
    Neighb51 = nodelist:mk_neighborhood([N66, N8, N50, N33, N42], N14, 3, 3),
    Neighb52 = nodelist:mk_neighborhood([N66, N8, N50, N14, N33, N42], N14, 3, 3),
    Neighb53 = nodelist:mk_neighborhood([N66, N8, N50, N14, N33, N42], N15, 3, 3),
    Neighb54 = nodelist:mk_neighborhood([N66, N8, N50, N33, N42], N14, 1, 1),
    Neighb55 = nodelist:mk_neighborhood([N66, N8, N50, N14, N33, N42], N14, 1, 1),
    Neighb56 = nodelist:mk_neighborhood([N66, N8, N50, N14, N33, N42], N15, 1, 1),
    ?compare_neighborhood(Neighb51, N14, N8, [N8, N66, N50], N33, [N33, N42, N50], true, true),
    ?compare_neighborhood(Neighb52, N14, N8, [N8, N66, N50], N33, [N33, N42, N50], true, true),
    ?compare_neighborhood(Neighb53, N15, N8, [N8, N66, N50], N33, [N33, N42, N50], true, true),
    ?compare_neighborhood(Neighb54, N14, N8, [N8], N33, [N33], true, true),
    ?compare_neighborhood(Neighb55, N14, N8, [N8], N33, [N33], true, true),
    ?compare_neighborhood(Neighb56, N15, N8, [N8], N33, [N33], true, true),
    ok.

%% @doc Tests truncating neighborhood structures.
-spec test_trunc(any()) -> ok.
test_trunc(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N50 = node:new(?PID(5), ?KEY(50), 0),
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20),
    
    Trunc11 = nodelist:trunc(Neighb1, 1, 1),
    Trunc12 = nodelist:trunc(Neighb1, 2, 1),
    Trunc13 = nodelist:trunc(Neighb1, 3, 1),
    Trunc14 = nodelist:trunc(Neighb1, 4, 1),
    Trunc15 = nodelist:trunc(Neighb1, 5, 1),
    Trunc16 = nodelist:trunc(Neighb1, 1, 2),
    Trunc17 = nodelist:trunc(Neighb1, 1, 3),
    Trunc18 = nodelist:trunc(Neighb1, 1, 4),
    Trunc19 = nodelist:trunc(Neighb1, 1, 5),
    Trunc110 = nodelist:trunc(Neighb1, 2, 2),
    Trunc111 = nodelist:trunc(Neighb1, 2, 3),
    Trunc112 = nodelist:trunc(Neighb1, 3, 2),
    Trunc113 = nodelist:trunc(Neighb1, 3, 3),
    Trunc114 = nodelist:trunc(Neighb1, 4, 4),
    Trunc115 = nodelist:trunc(Neighb1, 5, 5),
    ?compare_neighborhood(Trunc11,  N20, N14, [N14], N33, [N33], true, true),
    ?compare_neighborhood(Trunc12,  N20, N14, [N14, N8], N33, [N33], true, true),
    ?compare_neighborhood(Trunc13,  N20, N14, [N14, N8, N50], N33, [N33], true, true),
    ?compare_neighborhood(Trunc14,  N20, N14, [N14, N8, N50, N33], N33, [N33], true, true),
    ?compare_neighborhood(Trunc15,  N20, N14, [N14, N8, N50, N33], N33, [N33], true, true),
    ?compare_neighborhood(Trunc16,  N20, N14, [N14], N33, [N33, N50], true, true),
    ?compare_neighborhood(Trunc17,  N20, N14, [N14], N33, [N33, N50, N8], true, true),
    ?compare_neighborhood(Trunc18,  N20, N14, [N14], N33, [N33, N50, N8, N14], true, true),
    ?compare_neighborhood(Trunc19,  N20, N14, [N14], N33, [N33, N50, N8, N14], true, true),
    ?compare_neighborhood(Trunc110, N20, N14, [N14, N8], N33, [N33, N50], true, true),
    ?compare_neighborhood(Trunc111, N20, N14, [N14, N8], N33, [N33, N50, N8], true, true),
    ?compare_neighborhood(Trunc112, N20, N14, [N14, N8, N50], N33, [N33, N50], true, true),
    ?compare_neighborhood(Trunc113, N20, N14, [N14, N8, N50], N33, [N33, N50, N8], true, true),
    ?compare_neighborhood(Trunc114, N20, N14, [N14, N8, N50, N33], N33, [N33, N50, N8, N14], true, true),
    ?compare_neighborhood(Trunc115, N20, N14, [N14, N8, N50, N33], N33, [N33, N50, N8, N14], true, true),
    
    Trunc21 = nodelist:trunc_preds(Neighb1, 1),
    Trunc22 = nodelist:trunc_preds(Neighb1, 2),
    Trunc23 = nodelist:trunc_preds(Neighb1, 3),
    Trunc24 = nodelist:trunc_preds(Neighb1, 4),
    Trunc25 = nodelist:trunc_preds(Neighb1, 5),
    ?compare_neighborhood(Trunc21, N20, N14, [N14], N33, [N33, N50, N8, N14], true, true),
    ?compare_neighborhood(Trunc22, N20, N14, [N14, N8], N33, [N33, N50, N8, N14], true, true),
    ?compare_neighborhood(Trunc23, N20, N14, [N14, N8, N50], N33, [N33, N50, N8, N14], true, true),
    ?compare_neighborhood(Trunc24, N20, N14, [N14, N8, N50, N33], N33, [N33, N50, N8, N14], true, true),
    ?compare_neighborhood(Trunc25, N20, N14, [N14, N8, N50, N33], N33, [N33, N50, N8, N14], true, true),
    
    Trunc31 = nodelist:trunc_succs(Neighb1, 1),
    Trunc32 = nodelist:trunc_succs(Neighb1, 2),
    Trunc33 = nodelist:trunc_succs(Neighb1, 3),
    Trunc34 = nodelist:trunc_succs(Neighb1, 4),
    Trunc35 = nodelist:trunc_succs(Neighb1, 5),
    ?compare_neighborhood(Trunc31, N20, N14, [N14, N8, N50, N33], N33, [N33], true, true),
    ?compare_neighborhood(Trunc32, N20, N14, [N14, N8, N50, N33], N33, [N33, N50], true, true),
    ?compare_neighborhood(Trunc33, N20, N14, [N14, N8, N50, N33], N33, [N33, N50, N8], true, true),
    ?compare_neighborhood(Trunc34, N20, N14, [N14, N8, N50, N33], N33, [N33, N50, N8, N14], true, true),
    ?compare_neighborhood(Trunc35, N20, N14, [N14, N8, N50, N33], N33, [N33, N50, N8, N14], true, true),
    
    ok.

%% @doc Tests for nodelist:remove/2.
-spec test_remove2(any()) -> ok.
test_remove2(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), % N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N15 = node:new(?PID(2), ?KEY(15), 1), % N14 updated
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),

    % Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N8, Neighb1),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?compare_neighborhood(nodelist:remove(N14, Neighb1),
                          N33, N8, [N8, N50, N42], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:remove(N42, Neighb1),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:remove(N50, Neighb1),
                          N33, N14, [N14, N8, N42], N42, [N42, N8, N14], true, true),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(1), Neighb1),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb1),
                          N33, N8, [N8, N50, N42], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(5), Neighb1),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(6), Neighb1),
                          N33, N14, [N14, N8, N42], N42, [N42, N8, N14], true, true),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N9, Neighb1),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb1), Neighb1),
    ?equals(nodelist:remove(N66, Neighb1), Neighb1),
    ?equals(nodelist:remove(?PID(3), Neighb1), Neighb1),
    ?equals(nodelist:remove(?PID(7), Neighb1), Neighb1),

    % Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N8, Neighb2),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:remove(N14, Neighb2),
                          N33, N8, [N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:remove(N42, Neighb2),
                          N33, N14, [N14, N8, N50], N50, [N50, N8], true, true),
    ?compare_neighborhood(nodelist:remove(N50, Neighb2),
                          N33, N14, [N14, N8], N42, [N42, N8], true, true),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(1), Neighb2),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb2),
                          N33, N8, [N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(5), Neighb2),
                          N33, N14, [N14, N8, N50], N50, [N50, N8], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(6), Neighb2),
                          N33, N14, [N14, N8], N42, [N42, N8], true, true),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N9, Neighb2),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb2), Neighb2),
    ?equals(nodelist:remove(N66, Neighb2), Neighb2),
    ?equals(nodelist:remove(?PID(3), Neighb2), Neighb2),
    ?equals(nodelist:remove(?PID(7), Neighb2), Neighb2),

    % Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N8, Neighb3),
                          N33, N14, [N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:remove(N14, Neighb3),
                          N33, N8, [N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:remove(N42, Neighb3),
                          N33, N14, [N14, N8], N50, [N50], true, true),
    ?compare_neighborhood(nodelist:remove(N50, Neighb3),
                          N33, N14, [N14, N8], N42, [N42], true, true),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(1), Neighb3),
                          N33, N14, [N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb3),
                          N33, N8, [N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(5), Neighb3),
                          N33, N14, [N14, N8], N50, [N50], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(6), Neighb3),
                          N33, N14, [N14, N8], N42, [N42], true, true),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N9, Neighb3),
                          N33, N14, [N14], N42, [N42, N50], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb3), Neighb3),
    ?equals(nodelist:remove(N66, Neighb3), Neighb3),
    ?equals(nodelist:remove(?PID(3), Neighb3), Neighb3),
    ?equals(nodelist:remove(?PID(7), Neighb3), Neighb3),

    % Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N8, Neighb4),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:remove(N14, Neighb4),
                          N33, N42, [N42], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:remove(N42, Neighb4),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:remove(N50, Neighb4),
                          N33, N14, [N14], N42, [N42], true, true),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(1), Neighb4),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb4),
                          N33, N42, [N42], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(5), Neighb4),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:remove(?PID(6), Neighb4),
                          N33, N14, [N14], N42, [N42], true, true),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N9, Neighb4),
                          N33, N14, [N14], N42, [N42], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb4), Neighb4),
    ?equals(nodelist:remove(N66, Neighb4), Neighb4),
    ?equals(nodelist:remove(?PID(3), Neighb4), Neighb4),
    ?equals(nodelist:remove(?PID(7), Neighb4), Neighb4),

    % Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N14, Neighb5),
                          N33, N33, [N33], N33, [N33], false, false),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb5),
                          N33, N33, [N33], N33, [N33], false, false),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N15, Neighb5),
                          N33, N33, [N33], N33, [N33], false, false),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb5), Neighb5),
    ?equals(nodelist:remove(N9, Neighb5), Neighb5),
    ?equals(nodelist:remove(?PID(3), Neighb5), Neighb5),
    ?equals(nodelist:remove(?PID(1), Neighb5), Neighb5),

    % Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb6), Neighb6),
    ?equals(nodelist:remove(N9, Neighb6), Neighb6),
    ?equals(nodelist:remove(?PID(3), Neighb6), Neighb6),
    ?equals(nodelist:remove(?PID(1), Neighb6), Neighb6),

    ok.

%% @doc Tests for nodelist:remove/3.
-spec test_remove3(any()) -> ok.
test_remove3(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), % N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N15 = node:new(?PID(2), ?KEY(15), 1), % N14 updated
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),

    RemovedNodesTab = ets:new(test_remove3, [duplicate_bag, private]),
    AddNodeToTabFun = fun(Node) -> ets:insert(RemovedNodesTab, {Node}) end,
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),

    % Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N8, Neighb1, AddNodeToTabFun),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?equals(_N11 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N14, Neighb1, AddNodeToTabFun),
                          N33, N8, [N8, N50, N42], N42, [N42, N50, N8], true, true),
    ?equals(_N12 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N42, Neighb1, AddNodeToTabFun),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?equals(_N13 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N50, Neighb1, AddNodeToTabFun),
                          N33, N14, [N14, N8, N42], N42, [N42, N8, N14], true, true),
    ?equals(_N14 = ets:tab2list(RemovedNodesTab), [{N50}]),
    ets:delete_all_objects(RemovedNodesTab),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(1), Neighb1, AddNodeToTabFun),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?equals(_N15 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb1, AddNodeToTabFun),
                          N33, N8, [N8, N50, N42], N42, [N42, N50, N8], true, true),
    ?equals(_N16 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(5), Neighb1, AddNodeToTabFun),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?equals(_N17 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(6), Neighb1, AddNodeToTabFun),
                          N33, N14, [N14, N8, N42], N42, [N42, N8, N14], true, true),
    ?equals(_N18 = ets:tab2list(RemovedNodesTab), [{N50}]),
    ets:delete_all_objects(RemovedNodesTab),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N9, Neighb1, AddNodeToTabFun),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?equals(_N19 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb1, AddNodeToTabFun), Neighb1),
    ?equals(_N110 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(N66, Neighb1, AddNodeToTabFun), Neighb1),
    ?equals(_N111 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(3), Neighb1, AddNodeToTabFun), Neighb1),
    ?equals(_N112 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(7), Neighb1, AddNodeToTabFun), Neighb1),
    ?equals(_N113 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N8, Neighb2, AddNodeToTabFun),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),
    ?equals(_N21 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N14, Neighb2, AddNodeToTabFun),
                          N33, N8, [N8, N50], N42, [N42, N50, N8], true, true),
    ?equals(_N22 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N42, Neighb2, AddNodeToTabFun),
                          N33, N14, [N14, N8, N50], N50, [N50, N8], true, true),
    ?equals(_N23 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N50, Neighb2, AddNodeToTabFun),
                          N33, N14, [N14, N8], N42, [N42, N8], true, true),
    ?equals(_N24 = ets:tab2list(RemovedNodesTab), [{N50}]),
    ets:delete_all_objects(RemovedNodesTab),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(1), Neighb2, AddNodeToTabFun),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),
    ?equals(_N25 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb2, AddNodeToTabFun),
                          N33, N8, [N8, N50], N42, [N42, N50, N8], true, true),
    ?equals(_N26 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(5), Neighb2, AddNodeToTabFun),
                          N33, N14, [N14, N8, N50], N50, [N50, N8], true, true),
    ?equals(_N27 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(6), Neighb2, AddNodeToTabFun),
                          N33, N14, [N14, N8], N42, [N42, N8], true, true),
    ?equals(_N28 = ets:tab2list(RemovedNodesTab), [{N50}]),
    ets:delete_all_objects(RemovedNodesTab),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N9, Neighb2, AddNodeToTabFun),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),
    ?equals(_N29 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb2, AddNodeToTabFun), Neighb2),
    ?equals(_N210 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(N66, Neighb2, AddNodeToTabFun), Neighb2),
    ?equals(_N211 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(3), Neighb2, AddNodeToTabFun), Neighb2),
    ?equals(_N212 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(7), Neighb2, AddNodeToTabFun), Neighb2),
    ?equals(_N213 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N8, Neighb3, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42, N50], true, true),
    ?equals(_N31 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N14, Neighb3, AddNodeToTabFun),
                          N33, N8, [N8], N42, [N42, N50], true, true),
    ?equals(_N32 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N42, Neighb3, AddNodeToTabFun),
                          N33, N14, [N14, N8], N50, [N50], true, true),
    ?equals(_N33 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N50, Neighb3, AddNodeToTabFun),
                          N33, N14, [N14, N8], N42, [N42], true, true),
    ?equals(_N34 = ets:tab2list(RemovedNodesTab), [{N50}]),
    ets:delete_all_objects(RemovedNodesTab),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(1), Neighb3, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42, N50], true, true),
    ?equals(_N35 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb3, AddNodeToTabFun),
                          N33, N8, [N8], N42, [N42, N50], true, true),
    ?equals(_N36 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(5), Neighb3, AddNodeToTabFun),
                          N33, N14, [N14, N8], N50, [N50], true, true),
    ?equals(_N37 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(6), Neighb3, AddNodeToTabFun),
                          N33, N14, [N14, N8], N42, [N42], true, true),
    ?equals(_N38 = ets:tab2list(RemovedNodesTab), [{N50}]),
    ets:delete_all_objects(RemovedNodesTab),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N9, Neighb3, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42, N50], true, true),
    ?equals(_N39 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb3, AddNodeToTabFun), Neighb3),
    ?equals(_N310 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(N66, Neighb3, AddNodeToTabFun), Neighb3),
    ?equals(_N311 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(3), Neighb3, AddNodeToTabFun), Neighb3),
    ?equals(_N312 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(7), Neighb3, AddNodeToTabFun), Neighb3),
    ?equals(_N313 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N8, Neighb4, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42], true, true),
    ?equals(_N41 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N14, Neighb4, AddNodeToTabFun),
                          N33, N42, [N42], N42, [N42], true, true),
    ?equals(_N42 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N42, Neighb4, AddNodeToTabFun),
                          N33, N14, [N14], N14, [N14], true, true),
    ?equals(_N43 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(N50, Neighb4, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42], true, true),
    ?equals(_N44 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(1), Neighb4, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42], true, true),
    ?equals(_N45 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb4, AddNodeToTabFun),
                          N33, N42, [N42], N42, [N42], true, true),
    ?equals(_N46 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(5), Neighb4, AddNodeToTabFun),
                          N33, N14, [N14], N14, [N14], true, true),
    ?equals(_N47 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:remove(?PID(6), Neighb4, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42], true, true),
    ?equals(_N48 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N9, Neighb4, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42], true, true),
    ?equals(_N49 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb4, AddNodeToTabFun), Neighb4),
    ?equals(_N410 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(N66, Neighb4, AddNodeToTabFun), Neighb4),
    ?equals(_N411 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(3), Neighb4, AddNodeToTabFun), Neighb4),
    ?equals(_N412 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(7), Neighb4, AddNodeToTabFun), Neighb4),
    ?equals(_N413 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:remove(N14, Neighb5, AddNodeToTabFun),
                          N33, N33, [N33], N33, [N33], false, false),
    ?equals(_N51 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    % same with the node's pids:
    ?compare_neighborhood(nodelist:remove(?PID(2), Neighb5, AddNodeToTabFun),
                          N33, N33, [N33], N33, [N33], false, false),
    ?equals(_N52 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    % same with an updated node:
    ?compare_neighborhood(nodelist:remove(N15, Neighb5, AddNodeToTabFun),
                          N33, N33, [N33], N33, [N33], false, false),
    ?equals(_N53 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb5, AddNodeToTabFun), Neighb5),
    ?equals(_N510 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(N9, Neighb5, AddNodeToTabFun), Neighb5),
    ?equals(_N511 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(3), Neighb5, AddNodeToTabFun), Neighb5),
    ?equals(_N512 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(1), Neighb5, AddNodeToTabFun), Neighb5),
    ?equals(_N513 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    % remove non-existing nodes and pids:
    ?equals(nodelist:remove(N20, Neighb6, AddNodeToTabFun), Neighb6),
    ?equals(_N610 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(N9, Neighb6, AddNodeToTabFun), Neighb6),
    ?equals(_N611 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(3), Neighb6, AddNodeToTabFun), Neighb6),
    ?equals(_N612 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:remove(?PID(1), Neighb6, AddNodeToTabFun), Neighb6),
    ?equals(_N613 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    ets:delete(RemovedNodesTab),
    ok.

%% @doc Tests for nodelist:filter/2.
-spec test_filter2(any()) -> ok.
test_filter2(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), % N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),

    % Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N8) end),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N14) end),
                          N33, N8, [N8, N50, N42], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N42) end),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N50) end),
                          N33, N14, [N14, N8, N42], N42, [N42, N8, N14], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N20) end), Neighb1),
    ?equals(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N66) end), Neighb1),

    % Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N8) end),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N14) end),
                          N33, N8, [N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N42) end),
                          N33, N14, [N14, N8, N50], N50, [N50, N8], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N50) end),
                          N33, N14, [N14, N8], N42, [N42, N8], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N20) end), Neighb2),
    ?equals(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N66) end), Neighb2),

    % Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N8) end),
                          N33, N14, [N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N14) end),
                          N33, N8, [N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N42) end),
                          N33, N14, [N14, N8], N50, [N50], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N50) end),
                          N33, N14, [N14, N8], N42, [N42], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N20) end), Neighb3),
    ?equals(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N66) end), Neighb3),

    % Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N8) end),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N14) end),
                          N33, N42, [N42], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N42) end),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N50) end),
                          N33, N14, [N14], N42, [N42], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N20) end), Neighb4),
    ?equals(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N66) end), Neighb4),

    % Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb5, fun(N) -> not node:same_process(N, N14) end),
                          N33, N33, [N33], N33, [N33], false, false),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb5, fun(N) -> not node:same_process(N, N20) end), Neighb5),
    ?equals(nodelist:filter(Neighb5, fun(N) -> not node:same_process(N, N9) end), Neighb5),

    % Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb6, fun(N) -> not node:same_process(N, N20) end), Neighb6),
    ?equals(nodelist:filter(Neighb6, fun(N) -> not node:same_process(N, N9) end), Neighb6),

    ok.

%% @doc Tests for nodelist:filter/3.
-spec test_filter3(any()) -> ok.
test_filter3(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), % N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),

    RemovedNodesTab = ets:new(test_filter3, [duplicate_bag, private]),
    AddNodeToTabFun = fun(Node) -> ets:insert(RemovedNodesTab, {Node}) end,
    NodeListToTabListFun = fun(NodeList) -> lists:sort([{Node} || Node <- NodeList]) end,
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),

    % Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N8) end, AddNodeToTabFun),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?equals(_N11 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N14) end, AddNodeToTabFun),
                          N33, N8, [N8, N50, N42], N42, [N42, N50, N8], true, true),
    ?equals(_N12 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N42) end, AddNodeToTabFun),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?equals(_N13 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N50) end, AddNodeToTabFun),
                          N33, N14, [N14, N8, N42], N42, [N42, N8, N14], true, true),
    ?equals(_N14 = ets:tab2list(RemovedNodesTab), [{N50}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb1, fun(N) -> not lists:member(node:id(N), [?KEY(8), ?KEY(14)]) end, AddNodeToTabFun),
                          N33, N50, [N50, N42], N42, [N42, N50], true, true),
    ?equals(_N15 = lists:sort(ets:tab2list(RemovedNodesTab)), NodeListToTabListFun([N8, N14])),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N20) end, AddNodeToTabFun), Neighb1),
    ?equals(_N110 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:filter(Neighb1, fun(N) -> not node:same_process(N, N66) end, AddNodeToTabFun), Neighb1),
    ?equals(_N111 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N8) end, AddNodeToTabFun),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),
    ?equals(_N21 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N14) end, AddNodeToTabFun),
                          N33, N8, [N8, N50], N42, [N42, N50, N8], true, true),
    ?equals(_N22 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N42) end, AddNodeToTabFun),
                          N33, N14, [N14, N8, N50], N50, [N50, N8], true, true),
    ?equals(_N23 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N50) end, AddNodeToTabFun),
                          N33, N14, [N14, N8], N42, [N42, N8], true, true),
    ?equals(_N24 = ets:tab2list(RemovedNodesTab), [{N50}]),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N20) end, AddNodeToTabFun), Neighb2),
    ?equals(_N210 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:filter(Neighb2, fun(N) -> not node:same_process(N, N66) end, AddNodeToTabFun), Neighb2),
    ?equals(_N211 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N8) end, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42, N50], true, true),
    ?equals(_N31 = ets:tab2list(RemovedNodesTab), [{N8}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N14) end, AddNodeToTabFun),
                          N33, N8, [N8], N42, [N42, N50], true, true),
    ?equals(_N32 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N42) end, AddNodeToTabFun),
                          N33, N14, [N14, N8], N50, [N50], true, true),
    ?equals(_N33 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N50) end, AddNodeToTabFun),
                          N33, N14, [N14, N8], N42, [N42], true, true),
    ?equals(_N34 = ets:tab2list(RemovedNodesTab), [{N50}]),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N20) end, AddNodeToTabFun), Neighb3),
    ?equals(_N310 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:filter(Neighb3, fun(N) -> not node:same_process(N, N66) end, AddNodeToTabFun), Neighb3),
    ?equals(_N311 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N8) end, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42], true, true),
    ?equals(_N41 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N14) end, AddNodeToTabFun),
                          N33, N42, [N42], N42, [N42], true, true),
    ?equals(_N42 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N42) end, AddNodeToTabFun),
                          N33, N14, [N14], N14, [N14], true, true),
    ?equals(_N43 = ets:tab2list(RemovedNodesTab), [{N42}]),
    ets:delete_all_objects(RemovedNodesTab),
    ?compare_neighborhood(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N50) end, AddNodeToTabFun),
                          N33, N14, [N14], N42, [N42], true, true),
    ?equals(_N44 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N20) end, AddNodeToTabFun), Neighb4),
    ?equals(_N410 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:filter(Neighb4, fun(N) -> not node:same_process(N, N66) end, AddNodeToTabFun), Neighb4),
    ?equals(_N411 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter(Neighb5, fun(N) -> not node:same_process(N, N14) end, AddNodeToTabFun),
                          N33, N33, [N33], N33, [N33], false, false),
    ?equals(_N51 = ets:tab2list(RemovedNodesTab), [{N14}]),
    ets:delete_all_objects(RemovedNodesTab),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb5, fun(N) -> not node:same_process(N, N20) end, AddNodeToTabFun), Neighb5),
    ?equals(_N510 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:filter(Neighb5, fun(N) -> not node:same_process(N, N9) end, AddNodeToTabFun), Neighb5),
    ?equals(_N511 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    % Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    % remove non-existing nodes and pids:
    ?equals(nodelist:filter(Neighb6, fun(N) -> not node:same_process(N, N20) end, AddNodeToTabFun), Neighb6),
    ?equals(_N610 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),
    ?equals(nodelist:filter(Neighb6, fun(N) -> not node:same_process(N, N9) end, AddNodeToTabFun), Neighb6),
    ?equals(_N611 = ets:tab2list(RemovedNodesTab), []),
    ets:delete_all_objects(RemovedNodesTab),

    ets:delete(RemovedNodesTab),
    ok.

%% @doc Tests for nodelist:filter_min_length/4.
-spec test_filter_min_length4(any()) -> ok.
test_filter_min_length4(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),

    % Neighb1 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 0, 0),
                          N33, N14, [N14, N42], N42, [N42, N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 1, 1),
                          N33, N14, [N14, N42], N42, [N42, N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 2, 2),
                          N33, N14, [N14, N42], N42, [N42, N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 3, 3),
                          N33, N14, [N14, N8, N42], N42, [N42, N50, N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 4, 4),
                          N33, N14, [N14, N8, N50, N42], N42, [N42, N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 5, 5),
                          N33, N14, [N14, N8, N50, N42], N42, [N42, N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 2, 4),
                          N33, N14, [N14, N42], N42, [N42, N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 4, 2),
                          N33, N14, [N14, N8, N50, N42], N42, [N42, N14], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 0, 0), Neighb1),
    ?equals(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 3, 3), Neighb1),
    ?equals(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 2, 5), Neighb1),
    ?equals(nodelist:filter_min_length(Neighb1, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 5, 2), Neighb1),

    % Neighb2 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 3, 3),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 0, 0),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 1, 1),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 2, 2),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 3, 3),
                          N33, N14, [N14, N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 4, 4),
                          N33, N14, [N14, N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 2, 4),
                          N33, N14, [N14, N8], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 4, 2),
                          N33, N14, [N14, N8, N50], N42, [N42, N50], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 0, 0), Neighb2),
    ?equals(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 3, 3), Neighb2),
    ?equals(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 2, 5), Neighb2),
    ?equals(nodelist:filter_min_length(Neighb2, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 5, 2), Neighb2),

    % Neighb3 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 2, 2),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 0, 0),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 1, 1),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 2, 2),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 3, 3),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 2, 4),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N50)) end, 4, 2),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 0, 0), Neighb3),
    ?equals(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 3, 3), Neighb3),
    ?equals(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 2, 5), Neighb3),
    ?equals(nodelist:filter_min_length(Neighb3, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 5, 2), Neighb3),

    % Neighb4 = nodelist:mk_neighborhood([N8, N14, N42, N50], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter_min_length(Neighb4, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N42)) end, 0, 0),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb4, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N42)) end, 1, 1),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb4, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N42)) end, 2, 2),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb4, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N42)) end, 0, 1),
                          N33, N14, [N14], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb4, fun(N) -> not (node:same_process(N, N8) orelse node:same_process(N, N42)) end, 1, 0),
                          N33, N14, [N14], N14, [N14], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter_min_length(Neighb4, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 0, 0), Neighb4),
    ?equals(nodelist:filter_min_length(Neighb4, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 3, 3), Neighb4),
    ?equals(nodelist:filter_min_length(Neighb4, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 2, 5), Neighb4),
    ?equals(nodelist:filter_min_length(Neighb4, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 5, 2), Neighb4),

    % Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    % remove existing nodes:
    ?compare_neighborhood(nodelist:filter_min_length(Neighb5, fun(N) -> not node:same_process(N, N14) end, 0, 0),
                          N33, N33, [N33], N33, [N33], false, false),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb5, fun(N) -> not node:same_process(N, N14) end, 1, 1),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb5, fun(N) -> not node:same_process(N, N14) end, 0, 1),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:filter_min_length(Neighb5, fun(N) -> not node:same_process(N, N14) end, 1, 0),
                          N33, N14, [N14], N14, [N14], true, true),

    % remove non-existing nodes and pids:
    ?equals(nodelist:filter_min_length(Neighb5, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 0, 0), Neighb5),
    ?equals(nodelist:filter_min_length(Neighb5, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 3, 3), Neighb5),
    ?equals(nodelist:filter_min_length(Neighb5, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 2, 5), Neighb5),
    ?equals(nodelist:filter_min_length(Neighb5, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 5, 2), Neighb5),

    % Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    % remove non-existing nodes and pids:
    ?equals(nodelist:filter_min_length(Neighb6, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 0, 0), Neighb6),
    ?equals(nodelist:filter_min_length(Neighb6, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 3, 3), Neighb6),
    ?equals(nodelist:filter_min_length(Neighb6, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 2, 5), Neighb6),
    ?equals(nodelist:filter_min_length(Neighb6, fun(N) -> not (node:same_process(N, N20) orelse node:same_process(N, N66)) end, 5, 2), Neighb6),

    ok.

-spec test_merge(any()) -> ok.
test_merge(_Config) ->
    % use the same tests as in test_add_nodes/1 and let mk_beighborhood create the appropriate neighborhood structures:
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), % N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N15 = node:new(?PID(2), ?KEY(15), 1), % N14 updated
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N21 = node:new(?PID(3), ?KEY(21), 1), % N20 updated
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N50], N33),
    Neighb2 = nodelist:mk_neighborhood([N8, N14, N50], N33, 3, 3),
    Neighb3 = nodelist:mk_neighborhood([N8, N14, N50], N33, 2, 2),
    Neighb4 = nodelist:mk_neighborhood([N8, N14, N50], N33, 1, 1),
    Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    
    % Neighb1 = nodelist:mk_neighborhood([N8, N14, N50], N33),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N66], N42), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N42], N66), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66, N9], N20), 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66], N21), 3, 3),
                          N33, N21, [N21, N14, N8], N42, [N42, N50, N66], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N66], N42), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N42], N66), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66, N9], N20), 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N50, N66, N9], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66], N21), 4, 4),
                          N33, N21, [N21, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N66], N42), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N42], N66), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66, N9], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66], N21), 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N50], true, true),
    
    % Neighb2 = nodelist:mk_neighborhood([N8, N14, N50], N33, 3, 3),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N66], N42), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N42], N66), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66, N9], N20), 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66], N21), 3, 3),
                          N33, N21, [N21, N14, N8], N42, [N42, N50, N66], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N66], N42), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N42], N66), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66, N9], N20), 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N50, N66, N9], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66], N21), 4, 4),
                          N33, N21, [N21, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N66], N42), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N42], N66), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66, N9], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66], N21), 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N50], true, true),

    % Neighb3 = nodelist:mk_neighborhood([N8, N14, N50], N33, 2, 2),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N66], N42), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N42], N66), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66, N9], N20), 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66], N21), 3, 3),
                          N33, N21, [N21, N14, N8], N42, [N42, N50, N66], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N66], N42), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N42], N66), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66, N9], N20), 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N50, N66, N9], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66], N21), 4, 4),
                          N33, N21, [N21, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N66], N42), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N42], N66), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66, N9], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66], N21), 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N50], true, true),

    % Neighb4 = nodelist:mk_neighborhood([N8, N14, N50], N33, 1, 1),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N66], N42), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N42], N66), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66, N9], N20), 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66], N21), 3, 3),
                          N33, N21, [N21, N14, N66], N42, [N42, N50, N66], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66], N20), 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N66], N42), 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N42], N66), 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66, N9], N20), 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N50, N66, N9], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66], N21), 4, 4),
                          N33, N21, [N21, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N66], N42), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N42], N66), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66, N9], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66], N21), 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N50], true, true),
    
    % tests with (almost) empty successor/predecessor lists:
    % Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N66], N42), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N42], N66), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66, N9], N20), 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N66, N9], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66], N21), 3, 3),
                          N33, N21, [N21, N14, N66], N42, [N42, N66, N14], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66], N20), 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N66], N42), 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N42], N66), 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66, N9], N20), 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N66, N9, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66], N21), 4, 4),
                          N33, N21, [N21, N14, N66, N42], N42, [N42, N66, N14, N21], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N66], N42), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N42], N66), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66, N9], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66], N21), 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N66], true, true),
    
    % adding one node:
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:new_neighborhood(N20), 3, 3),
                          N33, N20, [N20, N14], N14, [N14, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:new_neighborhood(N15), 3, 3),
                          N33, N15, [N15], N15, [N15], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:new_neighborhood(N33), 3, 3),
                          N33, N14, [N14], N14, [N14], true, true),
    
%%     Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66], N20), 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N66], N42), 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N42], N66), 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66, N9], N20), 3, 3),
                          N33, N20, [N20, N9, N66], N42, [N42, N66, N9], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66], N21), 3, 3),
                          N33, N21, [N21, N66, N42], N42, [N42, N66, N21], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66], N20), 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N66], N42), 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N42], N66), 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66, N9], N20), 4, 4),
                          N33, N20, [N20, N9, N66, N42], N42, [N42, N66, N9, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66], N21), 4, 4),
                          N33, N21, [N21, N66, N42], N42, [N42, N66, N21], true, true),
    
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66], N20), 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N66], N42), 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N42], N66), 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N42, N66, N42], N20), 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66, N9], N20), 2, 2),
                          N33, N20, [N20, N9], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66], N21), 2, 2),
                          N33, N21, [N21, N66], N42, [N42, N66], true, true),
    
    % adding one node:
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:new_neighborhood(N20), 3, 3),
                          N33, N20, [N20], N20, [N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:new_neighborhood(N33), 3, 3),
                          N33, N33, [N33], N33, [N33], false, false),
    
    % some tests with the base node in the to-be-added nodes:
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N33, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N42, N66, N33], N20), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb1, nodelist:mk_neighborhood([N20, N42, N66], N33), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),

    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N33, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N42, N66, N33], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb2, nodelist:mk_neighborhood([N20, N42, N66], N33), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),

    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N33, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N42, N66, N33], N20), 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb3, nodelist:mk_neighborhood([N20, N42, N66], N33), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),

    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N33, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N42, N66, N33], N20), 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb4, nodelist:mk_neighborhood([N20, N42, N66], N33), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),

    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N33, N66], N20), 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N42, N66, N33], N20), 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb5, nodelist:mk_neighborhood([N20, N42, N66], N33), 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),

    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N33, N66], N20), 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N42, N66, N33], N20), 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:merge(Neighb6, nodelist:mk_neighborhood([N20, N42, N66], N33), 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    
    ok.

%% @doc Tests adding single nodes to a neighborhood structure.
-spec test_add_node(any()) -> ok.
test_add_node(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), % N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N15 = node:new(?PID(2), ?KEY(15), 1), % N14 updated
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N50], N33),
    Neighb2 = nodelist:mk_neighborhood([N8, N14, N50], N33, 3, 3),
    Neighb3 = nodelist:mk_neighborhood([N8, N14, N50], N33, 2, 2),
    Neighb4 = nodelist:mk_neighborhood([N8, N14, N50], N33, 1, 1),
    Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    
    % Neighb1 = nodelist:mk_neighborhood([N8, N14, N50], N33),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N20, 3, 3),
                          N33, N20, [N20, N14, N8], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N42, 3, 3),
                          N33, N14, [N14, N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N66, 3, 3),
                          N33, N14, [N14, N8, N66], N50, [N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N9, 3, 3),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N33, 3, 3),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb1, N20, 4, 4),
                          N33, N20, [N20, N14, N8, N50], N50, [N50, N8, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N42, 4, 4),
                          N33, N14, [N14, N8, N50, N42], N42, [N42, N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N66, 4, 4),
                          N33, N14, [N14, N8, N66, N50], N50, [N50, N66, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N9, 4, 4),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N33, 4, 4),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb1, N20, 2, 2),
                          N33, N20, [N20, N14], N50, [N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N42, 2, 2),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N66, 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N9, 2, 2),
                          N33, N14, [N14, N9], N50, [N50, N9], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb1, N33, 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N8], true, true),
    
    % Neighb2 = nodelist:mk_neighborhood([N8, N14, N50], N33, 3, 3),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N20, 3, 3),
                          N33, N20, [N20, N14, N8], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N42, 3, 3),
                          N33, N14, [N14, N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N66, 3, 3),
                          N33, N14, [N14, N8, N66], N50, [N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N9, 3, 3),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N33, 3, 3),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb2, N20, 4, 4),
                          N33, N20, [N20, N14, N8, N50], N50, [N50, N8, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N42, 4, 4),
                          N33, N14, [N14, N8, N50, N42], N42, [N42, N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N66, 4, 4),
                          N33, N14, [N14, N8, N66, N50], N50, [N50, N66, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N9, 4, 4),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N33, 4, 4),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb2, N20, 2, 2),
                          N33, N20, [N20, N14], N50, [N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N42, 2, 2),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N66, 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N9, 2, 2),
                          N33, N14, [N14, N9], N50, [N50, N9], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb2, N33, 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N8], true, true),

    % Neighb3 = nodelist:mk_neighborhood([N8, N14, N50], N33, 2, 2),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N20, 3, 3),
                          N33, N20, [N20, N14, N8], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N42, 3, 3),
                          N33, N14, [N14, N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N66, 3, 3),
                          N33, N14, [N14, N8, N66], N50, [N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N9, 3, 3),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N33, 3, 3),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb3, N20, 4, 4),
                          N33, N20, [N20, N14, N8, N50], N50, [N50, N8, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N42, 4, 4),
                          N33, N14, [N14, N8, N50, N42], N42, [N42, N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N66, 4, 4),
                          N33, N14, [N14, N8, N66, N50], N50, [N50, N66, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N9, 4, 4),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N33, 4, 4),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb3, N20, 2, 2),
                          N33, N20, [N20, N14], N50, [N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N42, 2, 2),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N66, 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N9, 2, 2),
                          N33, N14, [N14, N9], N50, [N50, N9], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb3, N33, 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N8], true, true),

    % Neighb4 = nodelist:mk_neighborhood([N8, N14, N50], N33, 1, 1),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N20, 3, 3),
                          N33, N20, [N20, N14, N50], N50, [N50, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N42, 3, 3),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N66, 3, 3),
                          N33, N14, [N14, N66, N50], N50, [N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N9, 3, 3),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N33, 3, 3),
                          N33, N14, [N14, N50], N50, [N50, N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb4, N20, 4, 4),
                          N33, N20, [N20, N14, N50], N50, [N50, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N42, 4, 4),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N66, 4, 4),
                          N33, N14, [N14, N66, N50], N50, [N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N9, 4, 4),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N33, 4, 4),
                          N33, N14, [N14, N50], N50, [N50, N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb4, N20, 2, 2),
                          N33, N20, [N20, N14], N50, [N50, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N42, 2, 2),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N66, 2, 2),
                          N33, N14, [N14, N66], N50, [N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N9, 2, 2),
                          N33, N14, [N14, N9], N50, [N50, N9], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb4, N33, 2, 2),
                          N33, N14, [N14, N50], N50, [N50, N14], true, true),
    
    % tests with (almost) empty successor/predecessor lists:
    % Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N20, 3, 3),
                          N33, N20, [N20, N14], N14, [N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N42, 3, 3),
                          N33, N14, [N14, N42], N42, [N42, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N66, 3, 3),
                          N33, N14, [N14, N66], N66, [N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N9, 3, 3),
                          N33, N14, [N14, N9], N9, [N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N33, 3, 3),
                          N33, N14, [N14], N14, [N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb5, N20, 4, 4),
                          N33, N20, [N20, N14], N14, [N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N42, 4, 4),
                          N33, N14, [N14, N42], N42, [N42, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N66, 4, 4),
                          N33, N14, [N14, N66], N66, [N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N9, 4, 4),
                          N33, N14, [N14, N9], N9, [N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N33, 4, 4),
                          N33, N14, [N14], N14, [N14], true, true),
    
    ?compare_neighborhood(nodelist:add_node(Neighb5, N20, 2, 2),
                          N33, N20, [N20, N14], N14, [N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N42, 2, 2),
                          N33, N14, [N14, N42], N42, [N42, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N66, 2, 2),
                          N33, N14, [N14, N66], N66, [N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N9, 2, 2),
                          N33, N14, [N14, N9], N9, [N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb5, N33, 2, 2),
                          N33, N14, [N14], N14, [N14], true, true),
    
%%     Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N20, 3, 3),
                          N33, N20, [N20], N20, [N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N42, 3, 3),
                          N33, N42, [N42], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N66, 3, 3),
                          N33, N66, [N66], N66, [N66], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N9, 3, 3),
                          N33, N9, [N9], N9, [N9], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N33, 3, 3),
                          N33, N33, [N33], N33, [N33], false, false),
    
    ?compare_neighborhood(nodelist:add_node(Neighb6, N20, 4, 4),
                          N33, N20, [N20], N20, [N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N42, 4, 4),
                          N33, N42, [N42], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N66, 4, 4),
                          N33, N66, [N66], N66, [N66], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N9, 4, 4),
                          N33, N9, [N9], N9, [N9], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N33, 4, 4),
                          N33, N33, [N33], N33, [N33], false, false),
    
    ?compare_neighborhood(nodelist:add_node(Neighb6, N20, 2, 2),
                          N33, N20, [N20], N20, [N20], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N42, 2, 2),
                          N33, N42, [N42], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N66, 2, 2),
                          N33, N66, [N66], N66, [N66], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N9, 2, 2),
                          N33, N9, [N9], N9, [N9], true, true),
    ?compare_neighborhood(nodelist:add_node(Neighb6, N33, 2, 2),
                          N33, N33, [N33], N33, [N33], false, false),
    
    % some tests with an updated base node in the to-be-added nodes:
    ?expect_exception(nodelist:add_node(nodelist:new_neighborhood(N14), N15, 1, 1),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),

    ok.

%% @doc Tests adding nodes to a neighborhood structure.
-spec test_add_nodes(any()) -> ok.
test_add_nodes(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), % N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N15 = node:new(?PID(2), ?KEY(15), 1), % N14 updated
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N21 = node:new(?PID(3), ?KEY(21), 1), % N20 updated
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N50], N33),
    Neighb2 = nodelist:mk_neighborhood([N8, N14, N50], N33, 3, 3),
    Neighb3 = nodelist:mk_neighborhood([N8, N14, N50], N33, 2, 2),
    Neighb4 = nodelist:mk_neighborhood([N8, N14, N50], N33, 1, 1),
    Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    
    % Neighb1 = nodelist:mk_neighborhood([N8, N14, N50], N33),
    ?equals(nodelist:add_nodes(Neighb1, [], 3, 3), Neighb1),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N42, N20, N66], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N66, N20, N42], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N9], 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N21], 3, 3),
                          N33, N21, [N21, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20], 3, 3),
                          N33, N20, [N20, N14, N8], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N42], 3, 3),
                          N33, N14, [N14, N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N66], 3, 3),
                          N33, N14, [N14, N8, N66], N50, [N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N9], 3, 3),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N33], 3, 3),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb1, [], 4, 4), Neighb1),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N42, N20, N66], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N66, N20, N42], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N9], 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N50, N66, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N21], 4, 4),
                          N33, N21, [N21, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20], 4, 4),
                          N33, N20, [N20, N14, N8, N50], N50, [N50, N8, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N42], 4, 4),
                          N33, N14, [N14, N8, N50, N42], N42, [N42, N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N66], 4, 4),
                          N33, N14, [N14, N8, N66, N50], N50, [N50, N66, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N9], 4, 4),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N33], 4, 4),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb1, [], 2, 2), nodelist:trunc(Neighb1, 2, 2)),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N42, N20, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N66, N20, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N9], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N21], 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20], 2, 2),
                          N33, N20, [N20, N14], N50, [N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N42], 2, 2),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N66], 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N9], 2, 2),
                          N33, N14, [N14, N9], N50, [N50, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N33], 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N8], true, true),
    
    % Neighb2 = nodelist:mk_neighborhood([N8, N14, N50], N33, 3, 3),
    ?equals(nodelist:add_nodes(Neighb2, [], 3, 3), Neighb2),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N42, N20, N66], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N66, N20, N42], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N9], 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N21], 3, 3),
                          N33, N21, [N21, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20], 3, 3),
                          N33, N20, [N20, N14, N8], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N42], 3, 3),
                          N33, N14, [N14, N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N66], 3, 3),
                          N33, N14, [N14, N8, N66], N50, [N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N9], 3, 3),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N33], 3, 3),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb2, [], 4, 4), Neighb2),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N42, N20, N66], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N66, N20, N42], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N9], 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N50, N66, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N21], 4, 4),
                          N33, N21, [N21, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20], 4, 4),
                          N33, N20, [N20, N14, N8, N50], N50, [N50, N8, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N42], 4, 4),
                          N33, N14, [N14, N8, N50, N42], N42, [N42, N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N66], 4, 4),
                          N33, N14, [N14, N8, N66, N50], N50, [N50, N66, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N9], 4, 4),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N33], 4, 4),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb2, [], 2, 2), nodelist:trunc(Neighb2, 2, 2)),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N42, N20, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N66, N20, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N9], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N21], 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20], 2, 2),
                          N33, N20, [N20, N14], N50, [N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N42], 2, 2),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N66], 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N9], 2, 2),
                          N33, N14, [N14, N9], N50, [N50, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N33], 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N8], true, true),

    % Neighb3 = nodelist:mk_neighborhood([N8, N14, N50], N33, 2, 2),
    ?equals(nodelist:add_nodes(Neighb3, [], 3, 3), Neighb3),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N42, N20, N66], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N66, N20, N42], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N9], 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N21], 3, 3),
                          N33, N21, [N21, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20], 3, 3),
                          N33, N20, [N20, N14, N8], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N42], 3, 3),
                          N33, N14, [N14, N8, N50], N42, [N42, N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N66], 3, 3),
                          N33, N14, [N14, N8, N66], N50, [N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N9], 3, 3),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N33], 3, 3),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb3, [], 4, 4), Neighb3),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N42, N20, N66], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N66, N20, N42], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N9], 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N50, N66, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N21], 4, 4),
                          N33, N21, [N21, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20], 4, 4),
                          N33, N20, [N20, N14, N8, N50], N50, [N50, N8, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N42], 4, 4),
                          N33, N14, [N14, N8, N50, N42], N42, [N42, N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N66], 4, 4),
                          N33, N14, [N14, N8, N66, N50], N50, [N50, N66, N8, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N9], 4, 4),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N33], 4, 4),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb3, [], 2, 2), Neighb3),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N42, N20, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N66, N20, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N9], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N21], 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20], 2, 2),
                          N33, N20, [N20, N14], N50, [N50, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N42], 2, 2),
                          N33, N14, [N14, N8], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N66], 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N9], 2, 2),
                          N33, N14, [N14, N9], N50, [N50, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N33], 2, 2),
                          N33, N14, [N14, N8], N50, [N50, N8], true, true),

    % Neighb4 = nodelist:mk_neighborhood([N8, N14, N50], N33, 1, 1),
    ?equals(nodelist:add_nodes(Neighb4, [], 3, 3), Neighb4),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N42, N20, N66], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N66, N20, N42], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N9], 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N21], 3, 3),
                          N33, N21, [N21, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20], 3, 3),
                          N33, N20, [N20, N14, N50], N50, [N50, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N42], 3, 3),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N66], 3, 3),
                          N33, N14, [N14, N66, N50], N50, [N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N9], 3, 3),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N33], 3, 3),
                          N33, N14, [N14, N50], N50, [N50, N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb4, [], 4, 4), Neighb4),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66], 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N42, N20, N66], 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N66, N20, N42], 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N9], 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N50, N66, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N21], 4, 4),
                          N33, N21, [N21, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20], 4, 4),
                          N33, N20, [N20, N14, N50], N50, [N50, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N42], 4, 4),
                          N33, N14, [N14, N50, N42], N42, [N42, N50, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N66], 4, 4),
                          N33, N14, [N14, N66, N50], N50, [N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N9], 4, 4),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N33], 4, 4),
                          N33, N14, [N14, N50], N50, [N50, N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb4, [], 2, 2), Neighb4),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N42, N20, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N66, N20, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N9], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N21], 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20], 2, 2),
                          N33, N20, [N20, N14], N50, [N50, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N42], 2, 2),
                          N33, N14, [N14, N50], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N66], 2, 2),
                          N33, N14, [N14, N66], N50, [N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N9], 2, 2),
                          N33, N14, [N14, N9], N50, [N50, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N33], 2, 2),
                          N33, N14, [N14, N50], N50, [N50, N14], true, true),
    
    % tests with (almost) empty successor/predecessor lists:
    % Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    ?equals(nodelist:add_nodes(Neighb5, [], 3, 3), Neighb5),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N42, N20, N66], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N66, N20, N42], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N9], 3, 3),
                          N33, N20, [N20, N14, N9], N42, [N42, N66, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N21], 3, 3),
                          N33, N21, [N21, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20], 3, 3),
                          N33, N20, [N20, N14], N14, [N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N42], 3, 3),
                          N33, N14, [N14, N42], N42, [N42, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N66], 3, 3),
                          N33, N14, [N14, N66], N66, [N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N9], 3, 3),
                          N33, N14, [N14, N9], N9, [N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N33], 3, 3),
                          N33, N14, [N14], N14, [N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb5, [], 4, 4), Neighb5),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66], 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N42, N20, N66], 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N66, N20, N42], 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N9], 4, 4),
                          N33, N20, [N20, N14, N9, N66], N42, [N42, N66, N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N21], 4, 4),
                          N33, N21, [N21, N14, N66, N42], N42, [N42, N66, N14, N21], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20], 4, 4),
                          N33, N20, [N20, N14], N14, [N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N42], 4, 4),
                          N33, N14, [N14, N42], N42, [N42, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N66], 4, 4),
                          N33, N14, [N14, N66], N66, [N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N9], 4, 4),
                          N33, N14, [N14, N9], N9, [N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N33], 4, 4),
                          N33, N14, [N14], N14, [N14], true, true),
    
    ?equals(nodelist:add_nodes(Neighb5, [], 2, 2), Neighb5),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N42, N20, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N66, N20, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N9], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N21], 2, 2),
                          N33, N21, [N21, N14], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20], 2, 2),
                          N33, N20, [N20, N14], N14, [N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N42], 2, 2),
                          N33, N14, [N14, N42], N42, [N42, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N66], 2, 2),
                          N33, N14, [N14, N66], N66, [N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N9], 2, 2),
                          N33, N14, [N14, N9], N9, [N9, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N33], 2, 2),
                          N33, N14, [N14], N14, [N14], true, true),
    
%%     Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    ?equals(nodelist:add_nodes(Neighb6, [], 3, 3), Neighb6),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66], 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N42, N20, N66], 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N66, N20, N42], 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N20, N42, N66, N42], 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N9], 3, 3),
                          N33, N20, [N20, N9, N66], N42, [N42, N66, N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N21], 3, 3),
                          N33, N21, [N21, N66, N42], N42, [N42, N66, N21], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20], 3, 3),
                          N33, N20, [N20], N20, [N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N42], 3, 3),
                          N33, N42, [N42], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N66], 3, 3),
                          N33, N66, [N66], N66, [N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N9], 3, 3),
                          N33, N9, [N9], N9, [N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N33], 3, 3),
                          N33, N33, [N33], N33, [N33], false, false),
    
    ?equals(nodelist:add_nodes(Neighb6, [], 4, 4), Neighb6),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66], 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N42, N20, N66], 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N66, N20, N42], 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N20, N42, N66, N42], 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N9], 4, 4),
                          N33, N20, [N20, N9, N66, N42], N42, [N42, N66, N9, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N21], 4, 4),
                          N33, N21, [N21, N66, N42], N42, [N42, N66, N21], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20], 4, 4),
                          N33, N20, [N20], N20, [N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N42], 4, 4),
                          N33, N42, [N42], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N66], 4, 4),
                          N33, N66, [N66], N66, [N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N9], 4, 4),
                          N33, N9, [N9], N9, [N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N33], 4, 4),
                          N33, N33, [N33], N33, [N33], false, false),
    
    ?equals(nodelist:add_nodes(Neighb6, [], 2, 2), Neighb6),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66], 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N42, N20, N66], 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N66, N20, N42], 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N20, N42, N66, N42], 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N9], 2, 2),
                          N33, N20, [N20, N9], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N21], 2, 2),
                          N33, N21, [N21, N66], N42, [N42, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20], 2, 2),
                          N33, N20, [N20], N20, [N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N42], 2, 2),
                          N33, N42, [N42], N42, [N42], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N66], 2, 2),
                          N33, N66, [N66], N66, [N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N9], 2, 2),
                          N33, N9, [N9], N9, [N9], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N33], 2, 2),
                          N33, N33, [N33], N33, [N33], false, false),
    
    % some tests with the base node in the to-be-added nodes:
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N33, N66], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N20, N42, N66, N33], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb1, [N33, N20, N42, N66], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),

    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N33, N66], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N20, N42, N66, N33], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb2, [N33, N20, N42, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),

    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N33, N66], 3, 3),
                          N33, N20, [N20, N14, N8], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N20, N42, N66, N33], 4, 4),
                          N33, N20, [N20, N14, N8, N66], N42, [N42, N50, N66, N8], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb3, [N33, N20, N42, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),

    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N33, N66], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N50, N66], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N20, N42, N66, N33], 4, 4),
                          N33, N20, [N20, N14, N66, N50], N42, [N42, N50, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb4, [N33, N20, N42, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N50], true, true),

    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N33, N66], 3, 3),
                          N33, N20, [N20, N14, N66], N42, [N42, N66, N14], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N20, N42, N66, N33], 4, 4),
                          N33, N20, [N20, N14, N66, N42], N42, [N42, N66, N14, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb5, [N33, N20, N42, N66], 2, 2),
                          N33, N20, [N20, N14], N42, [N42, N66], true, true),

    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N33, N66], 3, 3),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N20, N42, N66, N33], 4, 4),
                          N33, N20, [N20, N66, N42], N42, [N42, N66, N20], true, true),
    ?compare_neighborhood(nodelist:add_nodes(Neighb6, [N33, N20, N42, N66], 2, 2),
                          N33, N20, [N20, N66], N42, [N42, N66], true, true),
    
    % some tests with an updated base node in the to-be-added nodes:
    ?expect_exception(nodelist:add_nodes(nodelist:new_neighborhood(N14), [N15], 1, 1),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:add_nodes(nodelist:new_neighborhood(N14), [N8, N15], 1, 1),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:add_nodes(nodelist:new_neighborhood(N14), [N15, N8], 1, 1),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),

    ok.

%% @doc Tests adding nodes to a neighborhood structure.
-spec test_update_ids(any()) -> ok.
test_update_ids(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), % N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N15 = node:new(?PID(2), ?KEY(15), 1), % N14 updated
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N42 = node:new(?PID(5), ?KEY(42), 0),
    N50 = node:new(?PID(6), ?KEY(50), 0),
    N66 = node:new(?PID(7), ?KEY(66), 0),
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N50], N33),
    Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    
    % Neighb1 = nodelist:mk_neighborhood([N8, N14, N50], N33),
    ?equals(nodelist:add_nodes(Neighb1, [], 3, 3), Neighb1),
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N20, N42, N66]),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N20, N42, N66, N42]),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N20, N42, N66, N9]),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N20, N9, N66, N42]),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N14]),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N20]),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N9]),
                          N33, N14, [N14, N9, N50], N50, [N50, N9, N14], true, true),
    
    % tests with (almost) empty successor/predecessor lists:
    % Neighb5 = nodelist:mk_neighborhood([N14], N33, 1, 1),
    ?equals(nodelist:update_ids(Neighb5, []), Neighb5),
    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N20, N42, N66]),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N20, N42, N66, N42]),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N20, N42, N66, N15]),
                          N33, N15, [N15], N15, [N15], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N20, N15, N66, N42]),
                          N33, N15, [N15], N15, [N15], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N20]),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N14]),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N15]),
                          N33, N15, [N15], N15, [N15], true, true),
    
%%     Neighb6 = nodelist:mk_neighborhood([], N33, 1, 1),
    ?equals(nodelist:update_ids(Neighb6, []), Neighb6),
    ?compare_neighborhood(nodelist:update_ids(Neighb6, [N20, N42, N66]),
                          N33, N33, [N33], N33, [N33], false, false),
    ?compare_neighborhood(nodelist:update_ids(Neighb6, [N20, N42, N66, N42]),
                          N33, N33, [N33], N33, [N33], false, false),
    ?compare_neighborhood(nodelist:update_ids(Neighb6, [N20, N42, N66, N33]),
                          N33, N33, [N33], N33, [N33], false, false),
    ?compare_neighborhood(nodelist:update_ids(Neighb6, [N20]),
                          N33, N33, [N33], N33, [N33], false, false),
    
    % some tests with the base node in the to-be-updated nodes:
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N20, N42, N33, N66]),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N20, N42, N66, N33]),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb1, [N33, N20, N42, N66]),
                          N33, N14, [N14, N8, N50], N50, [N50, N8, N14], true, true),

    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N20, N42, N33, N66]),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N20, N42, N66, N33]),
                          N33, N14, [N14], N14, [N14], true, true),
    ?compare_neighborhood(nodelist:update_ids(Neighb5, [N33, N20, N42, N66]),
                          N33, N14, [N14], N14, [N14], true, true),

    ?compare_neighborhood(nodelist:update_ids(Neighb6, [N20, N42, N33, N66]),
                          N33, N33, [N33], N33, [N33], false, false),
    ?compare_neighborhood(nodelist:update_ids(Neighb6, [N20, N42, N66, N33]),
                          N33, N33, [N33], N33, [N33], false, false),
    ?compare_neighborhood(nodelist:update_ids(Neighb6, [N33, N20, N42, N66]),
                          N33, N33, [N33], N33, [N33], false, false),
    
    % some tests with an updated base node in the to-be-added nodes:
    ?expect_exception(nodelist:update_ids(nodelist:new_neighborhood(N14), [N15]),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:update_ids(nodelist:new_neighborhood(N14), [N8, N15]),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),
    ?expect_exception(nodelist:update_ids(nodelist:new_neighborhood(N14), [N15, N8]),
                      throw, 'cannot create a neighborhood() with a neighbor newer than the node itself'),

    ok.

%% @doc Tests converting neighborhood structures to node lists.
-spec test_to_list(any()) -> ok.
test_to_list(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N50 = node:new(?PID(5), ?KEY(50), 0),
    
    Neighb1 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20),
    Neighb2 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20, 3, 3),
    Neighb3 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20, 2, 2),
    Neighb4 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20, 1, 1),
    Neighb5 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20, 3, 1),
    Neighb6 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20, 3, 2),
    Neighb7 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20, 1, 3),
    Neighb8 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20, 2, 3),
    Neighb9 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20, 2, 1),
    Neighb10 = nodelist:mk_neighborhood([N8, N14, N33, N50], N20, 1, 2),
    
    ?equals(nodelist:to_list(Neighb1), [N20, N33, N50, N8, N14]),
    ?equals(nodelist:to_list(Neighb2), [N20, N33, N50, N8, N14]),
    ?equals(nodelist:to_list(Neighb3), [N20, N33, N50, N8, N14]),
    ?equals(nodelist:to_list(Neighb4), [N20, N33, N14]),
    ?equals(nodelist:to_list(Neighb5), [N20, N33, N50, N8, N14]),
    ?equals(nodelist:to_list(Neighb6), [N20, N33, N50, N8, N14]),
    ?equals(nodelist:to_list(Neighb7), [N20, N33, N50, N8, N14]),
    ?equals(nodelist:to_list(Neighb8), [N20, N33, N50, N8, N14]),
    ?equals(nodelist:to_list(Neighb9), [N20, N33, N8, N14]),
    ?equals(nodelist:to_list(Neighb10), [N20, N33, N50, N14]),
    
    ?equals(_N11 = nodelist:to_list(nodelist:new_neighborhood(N8)), [N8]),
    ?equals(_N12 = nodelist:to_list(nodelist:new_neighborhood(N8, N14)), [N8, N14]),
    ?equals(_N13 = nodelist:to_list(nodelist:new_neighborhood(N20, N8, N14)), [N8, N14, N20]),
    
    ok.

%% @doc Tests updating node IDs in two node lists.
-spec test_lupdate_ids(any()) -> ok.
test_lupdate_ids(_Config) ->
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), %N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N21 = node:new(?PID(3), ?KEY(21), 1), %N20 updated
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N50 = node:new(?PID(5), ?KEY(50), 0),

    ?equals(nodelist:lupdate_ids([N8, N14, N20, N33, N50], []), {[N8, N14, N20, N33, N50], []}),
    ?equals(nodelist:lupdate_ids([N8, N9, N14, N20, N33, N50], []), {[N9, N9, N14, N20, N33, N50], []}),
    ?equals(nodelist:lupdate_ids([N8, N9, N14, N20, N21, N33, N50], []), {[N9, N9, N14, N21, N21, N33, N50], []}),
    ?equals(nodelist:lupdate_ids([N8, N14, N20, N33], [N50]), {[N8, N14, N20, N33], [N50]}),
    ?equals(nodelist:lupdate_ids([N8, N14, N20], [N33, N50]), {[N8, N14, N20], [N33, N50]}),
    ?equals(nodelist:lupdate_ids([N8, N14], [N20, N33, N50]), {[N8, N14], [N20, N33, N50]}),
    ?equals(nodelist:lupdate_ids([N8], [N14, N20, N33, N50]), {[N8], [N14, N20, N33, N50]}),
    ?equals(nodelist:lupdate_ids([], [N8, N14, N20, N33, N50]), {[], [N8, N14, N20, N33, N50]}),
    ?equals(nodelist:lupdate_ids([], [N8, N9, N14, N20, N33, N50]), {[], [N9, N9, N14, N20, N33, N50]}),
    ?equals(nodelist:lupdate_ids([], [N8, N9, N14, N20, N21, N33, N50]), {[], [N9, N9, N14, N21, N21, N33, N50]}),
    
    ?equals(nodelist:lupdate_ids([N8, N14, N20, N33, N50], [N9]), {[N9, N14, N20, N33, N50], [N9]}),
    ?equals(nodelist:lupdate_ids([N8, N14, N20, N33, N50], [N8]), {[N8, N14, N20, N33, N50], [N8]}),
    ?equals(nodelist:lupdate_ids([N8, N14, N20, N33, N50], [N9, N33]), {[N9, N14, N20, N33, N50], [N9, N33]}),
    ?equals(nodelist:lupdate_ids([N8, N14, N20, N33, N50], [N8, N33]), {[N8, N14, N20, N33, N50], [N8, N33]}),
    ?equals(nodelist:lupdate_ids([N8, N14, N20, N33, N50], [N8, N14, N20, N33, N50]), {[N8, N14, N20, N33, N50], [N8, N14, N20, N33, N50]}),
    ?equals(nodelist:lupdate_ids([N8, N14, N21, N33, N50], [N9, N14, N20, N33, N50]), {[N9, N14, N21, N33, N50], [N9, N14, N21, N33, N50]}),
    ?equals(nodelist:lupdate_ids([N14, N21, N33, N50], [N9, N14, N20, N33]), {[N14, N21, N33, N50], [N9, N14, N21, N33]}),
    ?equals(nodelist:lupdate_ids([N14, N21, N33, N50], [N9, N14, N20, N33]), {[N14, N21, N33, N50], [N9, N14, N21, N33]}),
    
    ?equals(nodelist:lupdate_ids([N14, N20, N33, N50], [N9, N14, N20, N21, N33]), {[N14, N21, N33, N50], [N9, N14, N21, N21, N33]}),
    
    % some tests with "unsorted" lists:
    ?equals(nodelist:lupdate_ids([N50, N14, N21, N8, N33], [N33, N50, N9, N14, N20]), {[N50, N14, N21, N9, N33], [N33, N50, N9, N14, N21]}),
    ?equals(nodelist:lupdate_ids([N14, N33, N50, N21], [N9, N14, N20, N33]), {[N14, N33, N50, N21], [N9, N14, N21, N33]}),
    ?equals(nodelist:lupdate_ids([N14, N21, N33, N50], [N20, N9, N14, N33]), {[N14, N21, N33, N50], [N21, N9, N14, N33]}),
    ?equals(nodelist:lupdate_ids([N33, N14, N20, N50], [N9, N14, N20, N21, N33]), {[N33, N14, N21, N50], [N9, N14, N21, N21, N33]}),
    ok.

%% @doc Tests making removing outdated nodes from a node list.
-spec test_lremove_outdated(any()) -> ok.
test_lremove_outdated(_Config) ->
    Null = node:null(),
    N8 = node:new(?PID(1),   ?KEY(8), 0),
    N9 = node:new(?PID(1),   ?KEY(9), 1), %N8 updated
    N14 = node:new(?PID(2), ?KEY(14), 0),
    N20 = node:new(?PID(3), ?KEY(20), 0),
    N21 = node:new(?PID(3), ?KEY(21), 1), %N20 updated
    N33 = node:new(?PID(4), ?KEY(33), 0),
    N50 = node:new(?PID(5), ?KEY(50), 0),

    ?equals(nodelist:lremove_outdated([]), []),
    ?equals(nodelist:lremove_outdated([N8]), [N8]),
    ?equals(nodelist:lremove_outdated([N8, N14, N20, N33, N50]), [N8, N14, N20, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N9]), [N9]),
    ?equals(nodelist:lremove_outdated([N8, N9, N14, N20, N33, N50]), [N9, N14, N20, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N9, N14, N20, N21, N33, N50]), [N9, N14, N21, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N9, N14, N20, N8, N21, N33, N50]), [N9, N14, N21, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N9, N14, N20, N21, N33, N50, N50]), [N9, N14, N21, N33, N50, N50]),

    ?equals(nodelist:lremove_outdated([], N50), []),
    ?equals(nodelist:lremove_outdated([N8], N50), [N8]),
    ?equals(nodelist:lremove_outdated([N8], N9), []),
    ?equals(nodelist:lremove_outdated([N8], N50), [N8]),
    ?equals(nodelist:lremove_outdated([N8, N14, N20, N33], N50), [N8, N14, N20, N33]),
    ?equals(nodelist:lremove_outdated([N8, N14, N20, N33, N50], N50), [N8, N14, N20, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N14, N20, N33, N50], Null), [N8, N14, N20, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N14, N20, N33, N50], N21), [N8, N14, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N14, N20, N21, N33, N50], N21), [N8, N14, N21, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N9], N50), [N9]),
    ?equals(nodelist:lremove_outdated([N8, N9, N14, N20, N33, N50], N50), [N9, N14, N20, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N9, N14, N20, N21, N33, N50], N50), [N9, N14, N21, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N9, N14, N20, N8, N21, N33, N50], N50), [N9, N14, N21, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N9, N14, N20, N21, N33, N33, N50, N50], N50), [N9, N14, N21, N33, N33, N50, N50]),
    
    % some tests with "unsorted" lists:
    ?equals(nodelist:lremove_outdated([N14, N8, N20, N33, N50]), [N14, N8, N20, N33, N50]),
    ?equals(nodelist:lremove_outdated([N9, N8]), [N9]),
    ?equals(nodelist:lremove_outdated([N8, N20, N33, N50, N9, N14]), [N20, N33, N50, N9, N14]),
    ?equals(nodelist:lremove_outdated([N9, N14, N20, N8, N21, N33, N50]), [N9, N14, N21, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N9, N14, N20, N8, N21, N33, N50]), [N9, N14, N21, N33, N50]),
    ?equals(nodelist:lremove_outdated([N9, N50, N14, N20, N8, N21, N33, N50]), [N9, N50, N14, N21, N33, N50]),
    
    ?equals(nodelist:lremove_outdated([N14, N8, N20, N33, N50], N50), [N14, N8, N20, N33, N50]),
    ?equals(nodelist:lremove_outdated([N8, N20, N33, N14, N50], Null), [N8, N20, N33, N14, N50]),
    ?equals(nodelist:lremove_outdated([N8, N20, N33, N14, N50], N21), [N8, N33, N14, N50]),
    ?equals(nodelist:lremove_outdated([N8, N20, N33, N14, N21, N50], N21), [N8, N33, N14, N21, N50]),
    ?equals(nodelist:lremove_outdated([N9, N8], N50), [N9]),
    ?equals(nodelist:lremove_outdated([N8, N20, N33, N50, N9, N14], N50), [N20, N33, N50, N9, N14]),
    ?equals(nodelist:lremove_outdated([N9, N14, N20, N8, N21, N33, N50], N50), [N9, N14, N21, N33, N50]),
    ?equals(nodelist:lremove_outdated([N9, N14, N20, N8, N21, N33], N50), [N9, N14, N21, N33]),
    ?equals(nodelist:lremove_outdated([N9, N14, N20, N8, N21, N8, N33], N50), [N9, N14, N21, N33]),
    ?equals(nodelist:lremove_outdated([N9, N14, N20, N8, N21, N33], N50), [N9, N14, N21, N33]),
    ok.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% largest_smaller_than/2 and largest_smaller_than/3
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-spec prop_largest_smaller_than2(BaseNode::node:node_type(), AddNodes::[node:node_type()], PredsLength::pos_integer(), SuccsLength::pos_integer(), Id::?RT:key()) -> true.
prop_largest_smaller_than2(BaseNode, AddNodes_, PredsLength, SuccsLength, Id) ->
    % remove invalid and duplicate nodes from the list of nodes to add:
    AddNodes_2 = lists:usort(fun(A, B) -> node:pidX(A) =< node:pidX(B) end, AddNodes_),
    AddNodes = [Node || Node <- AddNodes_2,
                        not node:same_process(Node, BaseNode),
                        node:id(Node) =/= node:id(BaseNode)],
        
    prop_largest_smaller_than2_fixed_types(
      BaseNode, AddNodes, PredsLength, SuccsLength, Id).

-spec prop_largest_smaller_than2_fixed_types(BaseNode::node:node_type(), AddNodes::[node:node_type()], PredsLength::pos_integer(), SuccsLength::pos_integer(), Id::?RT:key()) -> true.
prop_largest_smaller_than2_fixed_types(BaseNode, AddNodes, PredsLength, SuccsLength, Id) ->
    % build new neighborhood
    Neighbors1 = nodelist:new_neighborhood(BaseNode),
    Neighbors2 = nodelist:add_nodes(Neighbors1, AddNodes, PredsLength, SuccsLength),
    
    LSTNode = nodelist:largest_smaller_than(Neighbors2, Id),
%%     ct:pal("nodelist:largest_smaller_than(~p, ~p)~n", [Neighbors2, Id]),
    case node:id(BaseNode) =:= Id andalso not nodelist:has_real_pred(Neighbors2) of
        true ->
            % in this case no node can be found
            ?equals_w_note(LSTNode, node:null(), "empty_neighbors");
        _ ->
            ?assert(node:is_valid(LSTNode)),
            ?assert(node:id(LSTNode) =/= Id),
            BetterNodes =
                [Node || Node <- [BaseNode | nodelist:to_list(Neighbors2)],
                         intervals:in(node:id(Node),
                                      intervals:new('(', node:id(LSTNode), Id, ')'))],
            ?equals([LSTNode | BetterNodes], [LSTNode])
    end.

tester_largest_smaller_than2(_Config) ->
    % some tests that previously failed:
    prop_largest_smaller_than2(
      {node, {{{197,151,185,165},1,{one,'ct@csr-pc40'}}, c,{}}, ?KEY(1), 0, 8000},
      [{node, {{{197,151,185,165},1,{one,'ct@csr-pc40'}}, c,{}}, ?KEY(3), 5, 8000},
       {node, {{{39,220,200,75},3,{two,'ct@csr-pc40'}}, c,101.41898888632392}, ?KEY(4), 0, 8000},
       {node, {{10,87,88,118},1,{three,'ct@csr-pc40'}}, ?KEY(5), 4, 8000}],
      1, 1, ?KEY(5)),
    
    prop_largest_smaller_than2(
      {node,
       {{6165,1898,39314,61608,61439,21180,27136,38528}, 4, {two,'ct@csr-pc40'}},
       ?KEY(0), 4, 8000},
      [{node,
        {{{38651,60988,24625,21004,23234,64811,37455,60050}, 5, {one,'ct@csr-pc40'}}, c,101.41898888632392},
        ?KEY(4), 1, 8000},
       {node,
        {{{12076,15918,45197,41953,9328,37838,58170,26800}, 4,one}, c,42},
        ?KEY(5), 0, 8000},
       {node, {{{25,8,53,113}, 5, {four,'ct@csr-pc40'}}, c, {3}}, ?KEY(2), 4, 8000}],
      1, 1, ?KEY(5)),
    
    tester:test(?MODULE, prop_largest_smaller_than2, 5, 5000, [{threads, 2}]).

-spec prop_largest_smaller_than3(BaseNode::node:node_type(), AddNodes::[node:node_type()], PredsLength::pos_integer(), SuccsLength::pos_integer(), Id::?RT:key()) -> true.
prop_largest_smaller_than3(BaseNode, AddNodes_, PredsLength, SuccsLength, Id) ->
    % remove invalid and duplicate nodes from the list of nodes to add:
    AddNodes_2 = lists:usort(fun(A, B) -> node:pidX(A) =< node:pidX(B) end, AddNodes_),
    AddNodes = [Node || Node <- AddNodes_2,
                        not node:same_process(Node, BaseNode),
                        node:id(Node) =/= node:id(BaseNode)],
        
    ?implies(AddNodes =/= [],
             prop_largest_smaller_than3_fixed_types(
               BaseNode, AddNodes, PredsLength, SuccsLength, Id, util:randomelem(AddNodes))).

-spec prop_largest_smaller_than3_fixed_types(BaseNode::node:node_type(), AddNodes::[node:node_type()], PredsLength::pos_integer(), SuccsLength::pos_integer(), Id::?RT:key(), LastFound::node:node_type()) -> true.
prop_largest_smaller_than3_fixed_types(BaseNode, AddNodes, PredsLength, SuccsLength, Id, LastFound) ->
    % build new neighborhood
    Neighbors1 = nodelist:new_neighborhood(BaseNode),
    Neighbors2 = nodelist:add_nodes(Neighbors1, AddNodes, PredsLength, SuccsLength),
    
    case node:id(LastFound) =:= Id of
        true -> true;
        _ ->
            LSTNode = nodelist:largest_smaller_than(Neighbors2, Id, LastFound),
%%             ct:pal("nodelist:largest_smaller_than(~p, ~p, ~p)~n", [Neighbors2, Id, LastFound]),
            case node:id(BaseNode) =:= Id andalso not nodelist:has_real_pred(Neighbors2) of
                true ->
                    % in this case no node can be found
                    ?equals_w_note(LSTNode, node:null(), "empty_neighbors");
                _ ->
%%                     ?assert(node:is_valid(LSTNode)), % always valid!
                    ?assert(node:id(LSTNode) =/= Id),
                    BetterNodes =
                        [Node || Node <- [BaseNode | nodelist:to_list(Neighbors2)],
                                 intervals:in(node:id(Node),
                                              intervals:new('(', node:id(LSTNode), Id, ')'))],
                    ?equals([LSTNode | BetterNodes], [LSTNode])
            end
    end.

tester_largest_smaller_than3(_Config) ->
    % some tests that previously failed:
    prop_largest_smaller_than3_fixed_types(
      {node, proc1, ?KEY(5), 5, 8000},
      [{node, proc2, ?KEY(0), 2, 8000}, {node, proc3, ?KEY(3), 2, 8000}],
      1, 3, ?KEY(5), {node, proc2, ?KEY(0), 2, 8000}),
    
    prop_largest_smaller_than3_fixed_types(
        {node, proc1, ?KEY(2), 1, 8000},
        [{node, proc2, ?KEY(5), 0, 8000}, {node, proc3, ?KEY(4), 4, 8000}],
        2, 5, ?KEY(3), {node, proc4, ?KEY(5), 0, 8000}),
    
    prop_largest_smaller_than3_fixed_types(
      {node, proc1, ?KEY(2), 1, 8000},
      [{node, proc2, ?KEY(5), 0, 8000}, {node, proc3, ?KEY(3), 1, 8000}],
      1, 1, ?KEY(2), {node, proc3, ?KEY(3), 1, 8000}),
    
    tester:test(?MODULE, prop_largest_smaller_than3, 5, 5000, [{threads, 2}]).

test_succ_ord_id(_Config) ->
    ?assert(nodelist:succ_ord_id(?KEY(5), ?KEY(6), ?KEY(3))),
    ?assert(nodelist:succ_ord_id(?KEY(6), ?KEY(5), ?KEY(3)) =:= false),
    ?assert(nodelist:succ_ord_id(?KEY(5), ?KEY(6), ?KEY(7))),
    ?assert(nodelist:succ_ord_id(?KEY(6), ?KEY(5), ?KEY(7)) =:= false),

    ?assert(nodelist:succ_ord_id(?KEY(5), ?KEY(5), ?KEY(5))),
    ?assert(nodelist:succ_ord_id(?KEY(5), ?KEY(5), ?KEY(6))),
    ?assert(nodelist:succ_ord_id(?KEY(5), ?KEY(5), ?KEY(4))),

    ?assert(nodelist:succ_ord_id(?KEY(8), ?KEY(4), ?KEY(5))),
    ?assert(nodelist:succ_ord_id(?KEY(8), ?MINUS_INFINITY, ?KEY(5))),
    ?assert(nodelist:succ_ord_id(?KEY(4), ?KEY(8), ?KEY(5)) =:= false),
    ?assert(nodelist:succ_ord_id(?MINUS_INFINITY, ?KEY(8), ?KEY(5)) =:= false),

    ?assert(nodelist:succ_ord_id(?KEY(4), ?KEY(5), ?MINUS_INFINITY)),
    ?assert(nodelist:succ_ord_id(?KEY(5), ?KEY(4), ?MINUS_INFINITY) =:= false),
    ?assert(nodelist:succ_ord_id(?KEY(5), ?MINUS_INFINITY, ?MINUS_INFINITY) =:= false),
    ?assert(nodelist:succ_ord_id(?MINUS_INFINITY, ?KEY(5), ?MINUS_INFINITY)),
    ?assert(nodelist:succ_ord_id(?MINUS_INFINITY, ?MINUS_INFINITY, ?MINUS_INFINITY)),

    ?assert(nodelist:succ_ord_id(?KEY(4), ?KEY(8), ?KEY(4))),
    ?assert(nodelist:succ_ord_id(?KEY(8), ?KEY(4), ?KEY(4)) =:= false).
